import math
import time
from typing import List, Union
from ..device import Compass
from ..pin import I2C, IOType

REG_XOUT_LSB = 0x00     # Output Data Registers for magnetic sensor.
REG_XOUT_MSB = 0x01
REG_YOUT_LSB = 0x02
REG_YOUT_MSB = 0x03
REG_ZOUT_LSB = 0x04
REG_ZOUT_MSB = 0x05
REG_STATUS_1 = 0x06     # Status Register.
REG_TOUT_LSB = 0x07     # Output Data Registers for temperature.
REG_TOUT_MSB = 0x08
REG_CONTROL_1 = 0x09    # Control Register #1.
REG_CONTROL_2 = 0x0a    # Control Register #2.
REG_RST_PERIOD = 0x0b   # SET/RESET Period Register.
REG_CHIP_ID = 0x0d      # Chip ID register.

# Flags for Status Register #1.
STAT_DRDY = 0b00000001  # Data Ready.
STAT_OVL = 0b00000010   # Overflow flag.
STAT_DOR = 0b00000100   # Data skipped for reading.

# Flags for Status Register #2.
INT_ENB = 0b00000001    # Interrupt Pin Enabling.
POL_PNT = 0b01000000    # Pointer Roll-over.
SOFT_RST = 0b10000000   # Soft Reset.

# Flags for Control Register 1.
MODE_STBY = 0b00000000  # Standby mode.
MODE_CONT = 0b00000001  # Continuous read mode.
ODR_10HZ = 0b00000000   # Output Data Rate Hz.
ODR_50HZ = 0b00000100
ODR_100HZ = 0b00001000
ODR_200HZ = 0b00001100
RNG_2G = 0b00000000     # Range 2 Gauss: for magnetic-clean environments.
RNG_8G = 0b00010000     # Range 8 Gauss: for strong magnetic fields.
OSR_512 = 0b00000000    # Over Sample Rate 512: less noise, more power.
OSR_256 = 0b01000000
OSR_128 = 0b10000000
OSR_64 = 0b11000000     # Over Sample Rate 64: more noise, less power.


class GY_271(Compass):
    # QMC5883L

    def __init__(self):
        super().__init__()
        self.i2c = I2C(0x0d)
        self.output_data_rate = ODR_100HZ
        self.output_range = RNG_8G
        self.oversampling_rate = OSR_512
        self.mode_cont = MODE_CONT | self.output_data_rate | self.output_range | self.oversampling_rate
        self.mode_stby = MODE_STBY | ODR_10HZ | RNG_2G | OSR_64

    def on_attached(self):
        assert any(p.type == IOType.SCL for p in self.pins) and any(
            p.type == IOType.SDA for p in self.pins), 'Invalid Pins'
        self.setup()

    def destory(self):
        self.set_standby()

    def setup(self):
        assert self.i2c.read_byte(REG_CHIP_ID) == 0xff,  'Fail to read chip id'
        self.set_continuous()

    def set_continuous(self):
        self.i2c.write_byte(REG_CONTROL_2, SOFT_RST)
        self.i2c.write_byte(REG_CONTROL_2, INT_ENB)
        self.i2c.write_byte(REG_RST_PERIOD, 0x01)
        self.i2c.write_byte(REG_CONTROL_1, self.mode_cont)

    def set_standby(self):
        self.i2c.write_byte(REG_CONTROL_2, SOFT_RST)
        self.i2c.write_byte(REG_CONTROL_2, INT_ENB)
        self.i2c.write_byte(REG_RST_PERIOD, 0x01)
        self.i2c.write_byte(REG_CONTROL_1, self.mode_stby)

    def read_word(self, reg: int):
        low = self.i2c.read_byte(reg)
        high = self.i2c.read_byte(reg + 1)
        val = (high << 8) + low
        return val

    def read_word_2c(self, reg: int):
        val = self.read_word(reg)
        if val >= 0x8000:  # 32768
            return val - 0x10000  # 65536
        else:
            return val

    def get_magnet_raw(self) -> List[float]:
        i = 0
        [x, y, z] = [None, None, None]
        while i < 20:  # Timeout after about 0.20 seconds.
            status = self.i2c.read_byte(REG_STATUS_1)
            if status & STAT_OVL:
                print('Magnetic sensor overflow')
                if self.output_range == RNG_2G:
                    print('Consider switching to RNG_8G output range')
            if status & STAT_DOR:
                # Previous measure was read partially, sensor in Data Lock.
                x = self.read_word_2c(REG_XOUT_LSB)
                y = self.read_word_2c(REG_YOUT_LSB)
                z = self.read_word_2c(REG_ZOUT_LSB)
                continue
            if status & STAT_DRDY:
                # Data is ready to read.
                x = self.read_word_2c(REG_XOUT_LSB)
                y = self.read_word_2c(REG_YOUT_LSB)
                z = self.read_word_2c(REG_ZOUT_LSB)
                break
            else:
                # Waiting for DRDY.
                time.sleep(0.01)
                i += 1
        return [x, y, z]

    def get_magnet(self) -> List[Union[float, None]]:
        [x, y, z] = self.get_magnet_raw()
        if x == None or y == None:
            [x1, y1] = [x, y]
        else:
            c = self.calibration
            x1 = x * c[0][0] + y * c[0][1] + c[0][2]
            y1 = x * c[1][0] + y * c[1][1] + c[1][2]
        return [x1, y1]

    def get_bearing_raw(self) -> Union[float, None]:
        [x, y, z] = self.get_magnet_raw()
        if x == None or y == None:
            return None

        b = math.degrees(math.atan2(y, x))
        if b < 0:
            b += 360.0
        return b

    def get_bearing(self) -> Union[float, None]:
        [x, y] = self.get_magnet()
        if x == None or y == None:
            return None

        b = math.degrees(math.atan2(y, x))
        if b < 0:
            b += 360.0
        b += self.declination
        if b < 0.0:
            b += 360.0
        elif b >= 360.0:
            b -= 360.0
        return b

    def detect(self,  timeout: float = 0.01) -> bool:
        value = self.get_bearing()
        if value:
            self.value = value
            return True
        return False



from ..pin import DIGPin
from ..device import Switch


class MicroSwitch(Switch):

    def __init__(self):
        super().__init__()
        self.pin: DIGPin = None

    def on_attached(self):
        self.pin = self.pins[0]
        assert self.pin, 'Invalid Pin'

    def read_state(self):
        return self.pin.input()



import smbus
import spidev
from abc import ABC, abstractmethod


class IOFlow():
    DEFAULT = 0
    IN = 1
    OUT = 2
    INOUT = 3


class IOType():
    GENERAL = 0
    PWM = 1
    ANA = 2
    SDA = 3
    SCL = 4
    MISO = 5
    MOSI = 6
    SCLK = 7


class Pin(ABC):

    def __init__(self, num: int, flow: IOFlow = IOFlow.OUT, type: IOType = IOType.GENERAL, tag: str = ''):
        super().__init__()
        self.num = num
        self.flow = flow
        self.type = type
        self.tag = tag

    def on_setup(self, host: 'Board'):
        pass

    def __eq__(self, other):
        return type(self) == type(other) and self.num == other.num and self.type == other.type

    def __str__(self):
        return 'Num:%s Flow:%s Type:%s' % (self.num, self.flow, self.type)


class DIGPin(Pin):

    @abstractmethod
    def input(self) -> int:
        pass

    @abstractmethod
    def output(self, val: int):
        pass


class ANAPin(Pin):

    @abstractmethod
    def input(self) -> float:
        pass

    @abstractmethod
    def output(self, val: float):
        pass


class PWMPin(Pin):

    def __init__(self, num: int, tag: str = ''):
        super().__init__(num, IOFlow.OUT, IOType.PWM, tag)

    @abstractmethod
    def start(self, dc: int = 0):
        pass

    @abstractmethod
    def stop(self):
        pass

    @abstractmethod
    def change_frequency(self, freq: int):
        pass

    @abstractmethod
    def change_duty_cycle(self, dc: int):
        pass


class SDAPin(Pin):

    def __init__(self, num: int, tag: str = ''):
        super().__init__(num, IOFlow.INOUT, IOType.SDA, tag)


class SCLPin(Pin):

    def __init__(self, num: int, tag: str = ''):
        super().__init__(num, IOFlow.DEFAULT, IOType.SCL, tag)


class MISOPin(Pin):

    def __init__(self, num: int, tag: str = ''):
        super().__init__(num, IOFlow.OUT, IOType.MISO, tag)


class MOSIPin(Pin):

    def __init__(self, num: int, tag: str = ''):
        super().__init__(num, IOFlow.IN, IOType.MOSI, tag)


class SCLKPin(Pin):

    def __init__(self, num: int, tag: str = ''):
        super().__init__(num, IOFlow.DEFAULT, IOType.SCLK, tag)


class I2C():

    def __init__(self, addr: int, bus: int = 1):
        super().__init__()
        self.addr = addr
        self.bus = smbus.SMBus(bus)

    def write_byte(self, reg: int, val: int):
        self.bus.write_byte_data(self.addr, reg, val)

    def read_byte(self, reg: int) -> int:
        return self.bus.read_byte_data(self.addr, reg)


class SPI():

    def __init__(self, bus: int = 0, client: int = 0):
        super().__init__()
        self.spi = spidev.SpiDev(bus, client)

    def set_mode(self, mode: int):
        self.spi.mode = mode

    def set_speed(self, hz: int):
        self.spi.max_speed_hz = hz

    def write_byte(self, val: int):
        self.spi.writebytes(val)

    def read_byte(self, reg: int) -> int:
        return self.spi.readbytes(reg)



from .GY_271 import GY_271
from .HAT import HAT, HATPin, HAT_SLOTS
from .HC_SR04 import HC_SR04
from .HP_ADDA import HP_ADDA, HP_ADPin, HP_ADDA_SLOTS
from .IRS_5 import IRS_5
from .KS_RGB import KS_RGB
from .LDR import LDR
from .MicroSwitch import MicroSwitch
from .QX_Eye import QX_Eye
from .Servo2Wire import Servo2Wire
from .SingleRelay import SingleRelay
from .RBP_34 import RBP_34, RBP_GPIO, RBP_PWM
from .SG90 import SG90



from typing import List
from ..device import CompoundEye
from ..pin import ANAPin

MAX_VOLT = 3.0


class QX_Eye(CompoundEye):

    def __init__(self):
        super().__init__()
        self.probe_angle = 360
        self.num_of_channels = 14
        self.min_val = 0
        self.max_val = 255
        self.val_pin: ANAPin = None
        self.dir_pin: ANAPin = None

    def on_attached(self):
        self.val_pin = self.pins[0]
        self.dir_pin = self.pins[1]
        assert self.val_pin and self.dir_pin, 'Invalid Pins'

    def get_val(self) -> int:
        val = self.val_pin.input()
        return min(val / MAX_VOLT * self.max_val, self.max_val)

    def get_dir(self) -> int:
        val = self.dir_pin.input()
        return min(val / MAX_VOLT * 360, 360)

    def detect(self) -> bool:
        self.value = self.max_ir = self.get_val()
        self.max_ir_dir = self.get_dir()
        return True



import signal
import subprocess
import sys
import time
import RPi.GPIO as GPIO

SWITCH_PIN = 26
GPIO.setmode(GPIO.BCM)
GPIO.setup(SWITCH_PIN, GPIO.IN, pull_up_down=GPIO.PUD_UP)


def main(args=sys.argv[1:]):
    while GPIO.input(SWITCH_PIN) != 0:
        time.sleep(0.05)
    start = time.time()
    proc = subprocess.Popen([sys.executable, args[0]])
    while proc.poll() is None:
        if GPIO.input(SWITCH_PIN) == 0 and time.time() - start > 1:
            break
        time.sleep(0.05)
    proc.send_signal(signal.SIGINT)
    time.sleep(0.5)
    proc.kill()
    main(args)


if __name__ == '__main__':
    main()



import time
from ..device import Board
from ..pin import ANAPin, IOFlow, IOType, MISOPin, MOSIPin, SCLKPin, SPI
from .RBP_34 import RBP_GPIO

GAIN = {'ADS1256_GAIN_1': 0,  # GAIN   1
        'ADS1256_GAIN_2': 1,  # GAIN   2
        'ADS1256_GAIN_4': 2,  # GAIN   4
        'ADS1256_GAIN_8': 3,  # GAIN   8
        'ADS1256_GAIN_16': 4,  # GAIN  16
        'ADS1256_GAIN_32': 5,  # GAIN  32
        'ADS1256_GAIN_64': 6,  # GAIN  64
        }

DRATE = {'ADS1256_30000SPS': 0xF0,  # reset the default values
         'ADS1256_15000SPS': 0xE0,
         'ADS1256_7500SPS': 0xD0,
         'ADS1256_3750SPS': 0xC0,
         'ADS1256_2000SPS': 0xB0,
         'ADS1256_1000SPS': 0xA1,
         'ADS1256_500SPS': 0x92,
         'ADS1256_100SPS': 0x82,
         'ADS1256_60SPS': 0x72,
         'ADS1256_50SPS': 0x63,
         'ADS1256_30SPS': 0x53,
         'ADS1256_25SPS': 0x43,
         'ADS1256_15SPS': 0x33,
         'ADS1256_10SPS': 0x20,
         'ADS1256_5SPS': 0x13,
         'ADS1256_2d5SPS': 0x03
         }

REG = {'REG_STATUS': 0,  # x1H
       'REG_MUX': 1,     # 01H
       'REG_ADCON': 2,   # 20H
       'REG_DRATE': 3,   # F0H
       'REG_IO': 4,      # E0H
       'REG_OFC0': 5,    # xxH
       'REG_OFC1': 6,    # xxH
       'REG_OFC2': 7,    # xxH
       'REG_FSC0': 8,    # xxH
       'REG_FSC1': 9,    # xxH
       'REG_FSC2': 10,   # xxH
       }

CMD = {'CMD_WAKEUP': 0x00,     # Completes SYNC and Exits Standby Mode 0000  0000 (00h)
       'CMD_RDATA': 0x01,      # Read Data 0000  0001 (01h)
       'CMD_RDATAC': 0x03,     # Read Data Continuously 0000   0011 (03h)
       # Stop Read Data Continuously 0000   1111 (0Fh)
       'CMD_SDATAC': 0x0F,
       'CMD_RREG': 0x10,       # Read from REG rrr 0001 rrrr (1xh)
       'CMD_WREG': 0x50,       # Write to REG rrr 0101 rrrr (5xh)
       # Offset and Gain Self-Calibration 1111    0000 (F0h)
       'CMD_SELFCAL': 0xF0,
       'CMD_SELFOCAL': 0xF1,   # Offset Self-Calibration 1111    0001 (F1h)
       'CMD_SELFGCAL': 0xF2,   # Gain Self-Calibration 1111    0010 (F2h)
       # System Offset Calibration 1111   0011 (F3h)
       'CMD_SYSOCAL': 0xF3,
       'CMD_SYSGCAL': 0xF4,    # System Gain Calibration 1111    0100 (F4h)
       # Synchronize the A/D Conversion 1111   1100 (FCh)
       'CMD_SYNC': 0xFC,
       'CMD_STANDBY': 0xFD,    # Begin Standby Mode 1111   1101 (FDh)
       'CMD_RESET': 0xFE,      # Reset to Power-Up Values 1111   1110 (FEh)
       }


class HP_ADPin(ANAPin):

    def __init__(self, num: int, tag=''):
        super().__init__(num, IOFlow.IN, IOType.ANA, tag)
        self.hpad: HP_ADDA = None

    def on_setup(self, host: Board):
        self.hpad = host

    def input(self) -> float:
        return self.hpad.get_value(self.num)

    def output(self, val: float):
        pass


HP_ADDA_SLOTS = [HP_ADPin(0), HP_ADPin(1), HP_ADPin(2), HP_ADPin(3),
                 HP_ADPin(4), HP_ADPin(5), HP_ADPin(6), HP_ADPin(7)]


class HP_ADDA(Board):
    # ADS1256

    RBP_PINS = (MOSIPin(9), MISOPin(10), SCLKPin(11),
                RBP_GPIO(17, IOFlow.IN, tag='drdy'),
                RBP_GPIO(18, tag='rst'), RBP_GPIO(22, tag='cs'))

    def __init__(self):
        super().__init__()
        self.spi = SPI()
        self.spi.set_mode(0b01)
        self.spi.set_speed(3000000)
        self.cs: RBP_GPIO = None
        self.drdy: RBP_GPIO = None
        self.rst: RBP_GPIO = None
        self.scan_mode = 0
        self.slots = HP_ADDA_SLOTS.copy()

    def on_attached(self):
        for pin in self.pins:
            setattr(self, pin.tag, pin)
        self.reset()
        self.wait_ready()
        self.setup(GAIN['ADS1256_GAIN_1'], DRATE['ADS1256_30000SPS'])

    def wait_ready(self):
        for i in range(0, 400000, 1):
            if(self.drdy.input() == 0):
                return
        raise Exception('HP_ADDA timeout')

    def setup(self, gain: int, drate: int):
        assert self.read_chip_id() == 3, 'Fail to read chip id'
        buf = [0, 0, 0, 0, 0, 0, 0, 0]
        buf[0] = (0 << 3) | (1 << 2) | (0 << 1)
        buf[1] = 0x08
        buf[2] = (0 << 5) | (0 << 3) | (gain << 0)
        buf[3] = drate
        self.cs.output(0)
        self.spi.write_byte([CMD['CMD_WREG'] | 0, 0x03])
        self.spi.write_byte(buf)
        self.cs.output(1)
        time.sleep(0.001)

    def reset(self):
        self.rst.output(1)
        time.sleep(0.2)
        self.rst.output(0)
        time.sleep(0.2)
        self.rst.output(1)

    def read_chip_id(self):
        id = self.read_data(REG['REG_STATUS'])
        id = id[0] >> 4
        return id

    def read_data(self, reg: int):
        self.cs.output(0)
        self.spi.write_byte([CMD['CMD_RREG'] | reg, 0x00])
        data = self.spi.read_byte(1)
        self.cs.output(1)
        return data

    def write_cmd(self, reg: int):
        self.cs.output(0)
        self.spi.write_byte([reg])
        self.cs.output(1)

    def write_reg(self, reg, data):
        self.cs.output(0)
        self.spi.write_byte([CMD['CMD_WREG'] | reg, 0x00, data])
        self.cs.output(1)

    def set_channel(self, channel):
        if channel > 7:
            return 0
        self.write_reg(REG['REG_MUX'], (channel << 4) | (1 << 3))

    def set_diff_channel(self, channel):
        if channel == 0:
            self.write_reg(REG['REG_MUX'], (0 << 4) | 1)
        elif channel == 1:
            self.write_reg(REG['REG_MUX'], (2 << 4) | 3)
        elif channel == 2:
            self.write_reg(REG['REG_MUX'], (4 << 4) | 5)
        elif channel == 3:
            self.write_reg(REG['REG_MUX'], (6 << 4) | 7)

    def read_adc_data(self):
        self.wait_ready()
        self.cs.output(0)
        self.spi.write_byte([CMD['CMD_RDATA']])
        buf = self.spi.read_byte(3)
        self.cs.output(1)
        read = (buf[0] << 16) & 0xff0000
        read |= (buf[1] << 8) & 0xff00
        read |= (buf[2]) & 0xff
        if (read & 0x800000):
            read &= 0xF000000
        return read

    def get_value(self, channel):
        value = 0
        if(self.scan_mode == 0):
            if(channel >= 8):
                return 0
            self.set_channel(channel)
            self.write_cmd(CMD['CMD_SYNC'])
            self.write_cmd(CMD['CMD_WAKEUP'])
            value = self.read_adc_data()
        else:
            if(channel >= 4):
                return 0
            self.set_diff_channel(channel)
            self.write_cmd(CMD['CMD_SYNC'])
            self.write_cmd(CMD['CMD_WAKEUP'])
            value = self.read_adc_data()
        return value*5.0/0x7fffff

    def get_values(self):
        values = [0, 0, 0, 0, 0, 0, 0, 0]
        for i in range(0, 8, 1):
            values[i] = self.get_value(i)
        return values



from ..device import Servo
from ..pin import PWMPin


class SG90(Servo):

    def __init__(self):
        super().__init__()
        self.pwm: PWMPin = None
        self.angle = 0

    def on_attached(self):
        self.pwm = self.pins[0]
        assert self.pwm, 'Invalid Pins'
        self.pwm.change_frequency(50)
        self.pwm.start()

    def move(self):
        assert self.angle >= 0 and self.angle <= 180, 'Invalid Angle'
        self.pwm.change_duty_cycle(2.5 + 10 * self.angle / 180)

    def stop(self):
        self.pos.stop()



import time
from ..device import Ultrasonic
from ..pin import DIGPin


class HC_SR04(Ultrasonic):

    def __init__(self):
        super().__init__()
        self.probe_angle = 15
        self.min_val = 2.0
        self.max_val = 450.0
        self.accuracy = 0.2
        self.trig: DIGPin = None
        self.echo: DIGPin = None

    def on_attached(self):
        for pin in self.pins:
            setattr(self, pin.tag, pin)
        assert self.trig and self.echo, 'Invalid Pins'

    def detect(self, timeout: float = 0.01) -> bool:
        self.trig.output(1)
        time.sleep(0.00001)
        self.trig.output(0)
        init = start = stop = time.time()
        while self.echo.input() == 0:
            start = time.time()
            if start - init > timeout:
                return False
        while self.echo.input() == 1:
            stop = time.time()
            if stop - init > timeout:
                return False
        elapsed = stop - start
        self.value = elapsed * 343 * 100 / 2
        return True



from typing import List
from ..device import CompoundEye

MAX_VOLT = 5.0


class IRS_5(CompoundEye):

    def __init__(self):
        super().__init__()
        self.probe_angle = 120
        self.num_of_channels = 5
        self.min_val = 0
        self.max_val = 255
        self.ir_vals: List[Tuple[int, int]] = []

    def on_attached(self):
        assert sum(p is not None for p in self.pins) > 0, 'Invalid Pins'

    def get_val(self, val: float) -> int:
        return min(val/MAX_VOLT*self.max_val, self.max_val)

    def detect(self) -> bool:
        self.ir_vals.clear()
        self.value = 0
        for pin in self.pins:
            if not pin:
                continue
            num = pin.num
            val = self.get_val(pin.input())
            if val > self.value:
                self.value = self.max_ir = val
                self.max_ir_dir = num
            self.ir_vals.append((num, val))
        return True




from ..pin import DIGPin, IOFlow, PWMPin, MISOPin, MOSIPin, SCLPin, SCLKPin, SDAPin
from ..device import Board, RCU
from ..engine import Engine
import RPi.GPIO as GPIO


class RBP_GPIO(DIGPin):

    def on_setup(self, host: Board):
        if self.flow == IOFlow.OUT:
            GPIO.setup(self.num, GPIO.OUT)
        elif self.flow == IOFlow.IN:
            if self.tag == 'pull_up':
                GPIO.setup(self.num, GPIO.IN, pull_up_down=GPIO.PUD_UP)
            elif self.tag == 'pull_down':
                GPIO.setup(self.num, GPIO.IN, pull_up_down=GPIO.PUD_DOWN)
            else:
                GPIO.setup(self.num, GPIO.IN)

    def input(self) -> int:
        assert self.flow == IOFlow.IN, 'Invalid IOFlow'
        return GPIO.input(self.num)

    def output(self, val: int):
        assert self.flow == IOFlow.OUT, 'Invalid IOFlow'
        GPIO.output(self.num, val)


class RBP_PWM(PWMPin):

    def on_setup(self, host: Board):
        GPIO.setup(self.num, GPIO.OUT)
        self.pwm = GPIO.PWM(self.num, 50)

    def start(self, dc=0):
        self.pwm.start(dc)

    def stop(self):
        self.pwm.stop()

    def change_frequency(self, freq):
        self.pwm.ChangeFrequency(freq)

    def change_duty_cycle(self, dc):
        self.pwm.ChangeDutyCycle(dc)


RBP_SLOTS = [SDAPin(2), SCLPin(3), RBP_GPIO(4), RBP_GPIO(5),
             RBP_GPIO(6), RBP_GPIO(7), RBP_GPIO(8), MOSIPin(9),
             MISOPin(10), SCLKPin(11), RBP_PWM(12), RBP_PWM(13),
             RBP_GPIO(14), RBP_GPIO(15), RBP_GPIO(16), RBP_GPIO(17),
             RBP_GPIO(18), RBP_GPIO(19), RBP_GPIO(20), RBP_GPIO(21),
             RBP_GPIO(22), RBP_GPIO(23), RBP_GPIO(24), RBP_GPIO(25)]


class RBP_34(RCU):

    def __init__(self):
        super().__init__()
        GPIO.setmode(GPIO.BCM)
        GPIO.setwarnings(False)
        self.num_of_cores = 4
        self.name = 'Raspberry Pi'
        self.slots = RBP_SLOTS.copy()

    def destory(self):
        super().destory()
        GPIO.cleanup()



from typing import Union
from ..device import Servo
from ..pin import DIGPin, PWMPin


class Servo2Wire(Servo):

    def __init__(self):
        super().__init__()
        self.prev_speed = self.speed
        self.prev_direction = self.direction
        self.pos: Union[PWMPin, DIGPin] = None
        self.neg: Union[PWMPin, DIGPin] = None

    def on_attached(self):
        for pin in self.pins:
            setattr(self, pin.tag, pin)
        assert self.pos and self.neg, 'Invalid Pins'
        self.pos.change_frequency(60)

    def on_detatched(self):
        self.stop()

    def move(self):
        self.speed <= 1.0, 'Invalid Speed'
        if self.speed != self.prev_speed:
            self.prev_speed = self.speed
            self.pos.change_duty_cycle(self.speed * 100)
        if self.direction != self.prev_direction:
            self.prev_direction = self.direction
            if self.direction == 1:
                self.pos.output(1)
                self.neg.output(0)
            elif self.direction == -1:
                self.pos.output(0)
                self.neg.output(1)

    def stop(self):
        self.pos.stop()



from setuptools import setup


setup(name='rbpros',
      version='0.10',
      description='ROS2 on Raspberry PI',
      url='',
      author='Dirk Zhou',
      author_email='dontbmh2011@gmail.com',
      license='MIT',
      packages=['rbpros', 'rbpros/suite'],
      install_requires=[
          'rpi.gpio',
          'smbus',
          'spidev'
      ],
      entry_points={
          'console_scripts': [
              'rbpros = rbpros.__main__:main',
          ]
      },
      python_requires='>=3.6',
      zip_safe=False)



from ..pin import ANAPin
from ..device import GrayScale


class LDR(GrayScale):

    def __init__(self):
        super().__init__()
        self.ctr = 0.6
        self.pin: ANAPin = None

    def on_attached(self):
        self.pin = self.pins[0]
        assert self.pin, 'Invalid Pin'

    def detect(self) -> bool:
        v = self.pin.input()
        self.value = self.normalize(v)
        return True

    def normalize(self, val: float) -> float:
        min_val = self.max_val * self.ctr
        return max(0, (val - min_val) / 100.0 * self.max_val)



from ..device import Board
from ..pin import DIGPin, I2C, IOType, PWMPin, SCLPin, SDAPin
import math
import time

SUBADR1 = 0x02
SUBADR2 = 0x03
SUBADR3 = 0x04
MODE1 = 0x00
PRESCALE = 0xFE
LED0_ON_L = 0x06
LED0_ON_H = 0x07
LED0_OFF_L = 0x08
LED0_OFF_H = 0x09
ALLLED_ON_L = 0xFA
ALLLED_ON_H = 0xFB
ALLLED_OFF_L = 0xFC
ALLLED_OFF_H = 0xFD


class HATPin(PWMPin, DIGPin):

    def __init__(self, num: int, tag=''):
        super().__init__(num, tag)
        self.pwm: int = 0
        self.hat: HAT = None

    def on_setup(self, host: Board):
        if self.num == 1 or self.num == 2:
            self.pwm = 0
        if self.num == 3 or self.num == 4:
            self.pwm = 5
        self.hat = host

    def input(self) -> int:
        pass

    def output(self, val: int):
        self.hat.set_level(self.num, val)

    def start(self, dc=0):
        self.change_duty_cycle(dc)

    def stop(self):
        self.change_duty_cycle(0)

    def change_frequency(self, freq):
        self.hat.set_frequency(freq)

    def change_duty_cycle(self, dc):
        self.hat.set_duty_cycle(self.pwm, dc)


HAT_SLOTS = [HATPin(1), HATPin(2), HATPin(3), HATPin(4)]


class HAT(Board):
    # PCA9685

    RBP_PINS = (SDAPin(2), SCLPin(3))

    def __init__(self):
        super().__init__()
        self.i2c = I2C(0x40)
        self.slots = HAT_SLOTS.copy()

    def on_attached(self):
        assert any(p.type == IOType.SCL for p in self.pins) and any(
            p.type == IOType.SDA for p in self.pins), 'Invalid Pins'
        self.i2c.write_byte(MODE1, 0x00)

    def set_frequency(self, freq: int):
        prescale = 25000000.0    # 25MHz
        prescale /= 4096.0       # 12-bit
        prescale /= float(freq)
        prescale -= 1.0
        prescale = math.floor(prescale + 0.5)
        oldmode = self.i2c.read_byte(MODE1)
        newmode = (oldmode & 0x7F) | 0x10        # sleep
        self.i2c.write_byte(MODE1, newmode)        # go to sleep
        self.i2c.write_byte(PRESCALE, prescale)
        self.i2c.write_byte(MODE1, oldmode)
        time.sleep(0.005)
        self.i2c.write_byte(MODE1, oldmode | 0x80)

    def set_pwm(self, channel: int, on: int, off: int):
        self.i2c.write_byte(LED0_ON_L + 4*channel, on & 0xFF)
        self.i2c.write_byte(LED0_ON_H + 4*channel, on >> 8)
        self.i2c.write_byte(LED0_OFF_L + 4*channel, off & 0xFF)
        self.i2c.write_byte(LED0_OFF_H + 4*channel, off >> 8)

    def set_duty_cycle(self, channel: int, pulse: int):
        self.set_pwm(channel, 0, int(pulse * (4095 / 100)))

    def set_level(self, channel: int, value: int):
        if (value == 1):
            self.set_pwm(channel, 0, 4095)
        else:
            self.set_pwm(channel, 0, 0)
        time.sleep(0.05)



from rclpy.executors import Task
from rclpy.node import Node, Publisher, Subscription
from typing import Callable, Coroutine, Dict, Generic, List, Type, TypeVar, Union
from . import device
from .util import generate_id
import time

T = TypeVar('T')
TB = TypeVar('TB', bound='Behavior')
TD = TypeVar('TD', bound='Device')


class Behavior(Node):

    def __init__(self):
        id = generate_id(type(self))
        super().__init__(id)
        self.id = id
        self._enabled = True
        self._engine = None
        self._inited = False
        self._update = getattr(self, 'on_update')

    @property
    def enabled(self) -> bool:
        return self._enabled

    @enabled.setter
    def enabled(self, val: bool):
        if self._enabled == val:
            return
        self._enabled = val
        if self._enabled:
            self.on_enable()
        else:
            self.on_disable()

    def run_task(self, callback: Union[Callable, Coroutine]) -> Task:
        return self._engine.execute(callback, isinstance(callback, Coroutine))

    async def delay(self, sec: float):
        if sec == 0:
            await self.run_task(lambda: None)
        else:
            start = time.time()
            while time.time() - start < sec:
                await self.run_task(lambda: None)

    def on_awake(self):
        pass

    def on_start(self):
        pass

    def on_enable(self):
        pass

    def on_disable(self):
        pass

    def on_update(self):
        pass

    def on_destory(self):
        pass


class EventBehavior(Behavior):

    def __init__(self):
        super().__init__()
        self._pubrs: Dict[str, Publisher] = {}
        self._subrs: Dict[Callable, List[Subscription]] = {}

    def publish(self, topic: str, msg_type: Type[T], msg: T):
        pubr = self._pubrs.get(topic)
        if not pubr:
            pubr = self.create_publisher(msg_type, topic, 10)
            self._pubrs[topic] = pubr
        pubr.publish(msg)

    def subscribe(self, topic: str, msg_type: Type[T], handler: Callable[[T], None]):
        subrs = self._subrs.get(handler)
        if not subrs:
            self._subrs[handler] = subrs = []
        subr = self.create_subscription(msg_type, topic, handler, 10)
        subrs.append(subr)

    def unsubscribe(self, handler: Callable):
        subrs = self._subrs.get(handler)
        if subrs:
            for subr in subrs:
                self.destroy_subscription(subr)
            self._subrs.pop(handler)


class DeviceBehavior(EventBehavior, Generic[TD]):

    def __init__(self):
        super().__init__()
        self.device: U = None

    @property
    def device_name(self) -> str:
        return self.device.name

    def attach_behavior(self, klass: Type[TB]) -> TB:
        return self.device.attach_behavior(klass)

    def detatch_behavior(self, beh: Behavior):
        self.device.detatch_behavior(beh)

    def get_behavior(self, klass: Type[TB], include_host: bool = False, include_attachments: bool = False) -> TB:
        beh = self.device.get_behavior(klass)
        if not beh and include_host and isinstance(self.device, device.Attachable):
            beh = self.device.get_behavior_include_host(klass)
        if not beh and include_attachments and isinstance(self.device, device.Board):
            beh = self.device.get_behavior_include_attachments(klass)
        return beh

    def get_behaviors(self, klass: Type[TB], include_host: bool = False, include_attachments: bool = False) -> List[TB]:
        behs = []
        behs.extend(self.device.get_behaviors(klass))
        if include_host and isinstance(self.device, device.Attachable):
            behs.extend(self.device.get_behaviors_include_host(klass))
        if include_attachments and isinstance(self.device, device.Board):
            behs.extend(self.device.get_behaviors_include_attachments(klass))
        return behs



from typing import List
from ..pin import PWMPin
from ..device import Bulb


class KS_RGB(Bulb):

    def __init__(self):
        super().__init__()
        self.red: PWMPin = None
        self.green: PWMPin = None
        self.blue: PWMPin = None

    def on_attached(self):
        for pin in self.pins:
            setattr(self, pin.tag, pin)
            pin.change_frequency(70)
            pin.start()

    def change_color(self, color: List[int]):
        if self.red:
            self.red.change_duty_cycle((int)(color[0]/255*100))
        if self.green:
            self.green.change_duty_cycle((int)(color[1]/255*100))
        if self.blue:
            self.blue.change_duty_cycle((int)(color[2]/255*100))



from ..pin import DIGPin
from ..device import Relay


class SingleRelay(Relay):

    def __init__(self):
        super().__init__()
        self.pin: DIGPin = None

    def on_attached(self):
        self.pin = self.pins[0]
        assert self.pin, 'Invalid Pin'

    def set_state(self, on: int):
        self.pin.output(on)
